/*
 * --- GSMP-COPYRIGHT-NOTE-BEGIN ---
 * 
 * This copyright note is auto-generated by ./scripts/Create-CopyPatch.
 * Please add additional copyright information _after_ the line containing
 * the GSMP-COPYRIGHT-NOTE-END tag. Otherwise it might get removed by
 * the ./scripts/Create-CopyPatch script. Do not edit this copyright text!
 * 
 * GSMP: utility/tests/Attribute.cc
 * General Sound Manipulation Program is Copyright (C) 2000 - 2004
 *   Valentin Ziegler and René Rebe
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2. A copy of the GNU General
 * Public License can be found in the file LICENSE.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-
 * ABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * --- GSMP-COPYRIGHT-NOTE-END ---
 */

#include "Attribute.hh"

class foo
{
 public:
  const Utility::Attribute<int>& bar;

  foo ();

 private:
  Utility::Attribute<int> real_bar;
};

foo::foo ()
  : bar (real_bar)
{
}

int main ()
{
  Utility::Attribute<int, true> a_with_signal;
  Utility::Attribute<int, false> a_without_signal;
  int x;
  x = a_with_signal.Get ();
  x = a_without_signal.Get ();

  foo f;
  f.bar.Get ();

  return 0;
}
