/*
 * --- GSMP-COPYRIGHT-NOTE-BEGIN ---
 * 
 * This copyright note is auto-generated by ./scripts/Create-CopyPatch.
 * Please add additional copyright information _after_ the line containing
 * the GSMP-COPYRIGHT-NOTE-END tag. Otherwise it might get removed by
 * the ./scripts/Create-CopyPatch script. Do not edit this copyright text!
 * 
 * GSMP: utility/include/Logging.hh
 * General Sound Manipulation Program is Copyright (C) 2000 - 2004
 *   Valentin Ziegler and René Rebe
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2. A copy of the GNU General
 * Public License can be found in the file LICENSE.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-
 * ABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * --- GSMP-COPYRIGHT-NOTE-END ---
 */

#ifndef UTILITY__LOGGING_HH__
#define UTILITY__LOGGING_HH__

#include "Logger.hh"
#include "config/DebugConfig.hh"

namespace Utility
{
  extern UtilityLogDevice UtilityLogDev;
  extern UtilityLogDestination UtilityLogDst;

  template <typename WL_TRAITS>
  class UtilityLogger :
    public Logger <UtilityLogDestinationConfig,
		   UtilityLogDeviceConfig,
		   WL_TRAITS>
  {
  public:
    UtilityLogger () : Logger <UtilityLogDestinationConfig,
			       UtilityLogDeviceConfig,
			       WL_TRAITS> (UtilityLogDst) {}
  };

  template <typename WL_TRAITS, typename OBJ>
  class UtilityObjectLogger :
    public ObjectLogger <UtilityLogDestinationConfig,
			 UtilityLogDeviceConfig,
			 WL_TRAITS, OBJ>
  {
  public:
    UtilityObjectLogger (OBJ* object) :
      ObjectLogger <UtilityLogDestinationConfig,
		    UtilityLogDeviceConfig,
		    WL_TRAITS, OBJ> (UtilityLogDst, object) {}
  };
  
  extern UtilityLogger <DefaultLogging_Utility> UtilityLog;
  
} // end namespace Utility

#endif // UTILITY__LOGGING_HH__
