/*
 * --- GSMP-COPYRIGHT-NOTE-BEGIN ---
 * 
 * This copyright note is auto-generated by ./scripts/Create-CopyPatch.
 * Please add additional copyright information _after_ the line containing
 * the GSMP-COPYRIGHT-NOTE-END tag. Otherwise it might get removed by
 * the ./scripts/Create-CopyPatch script. Do not edit this copyright text!
 * 
 * GSMP: utility/include/template/Storage.tcc
 * General Sound Manipulation Program is Copyright (C) 2000 - 2004
 *   Valentin Ziegler and René Rebe
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2. A copy of the GNU General
 * Public License can be found in the file LICENSE.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-
 * ABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * --- GSMP-COPYRIGHT-NOTE-END ---
 */

#include "Compiler.hh"

#if GCC_VERSION < 30300 || __INCLUDE_LEVEL__ == 1

#ifndef UTILITY__STORAGE_TCC__
#define UTILITY__STORAGE_TCC__

namespace Utility {
  
  template <> std::istream& Storage<bool>::Read (std::istream& is) {
    // TODO: error handling
    is >> value;
    if (is.fail () ) {
      
      is.clear ();
      
      std::string s;
      is >> s;
      
      if (s == "true")
	value = true;
      else if (s == "false")
	value = false;
      else 
	std::cout << "Error: Unable to read boolean value for key \""
		  << name << "\"!";
    }
    return is;
  }

} // namespace Utility

#endif // UTILITY__STORAGE_TCC__

#endif

